<?php

/**
 * CreepTD is an Open Source Online Multiplayer TowerDefense game for 1-4
 * players playing on one map. It's the mother of all TD games.
 *
 *                      ---   www.creeptd.com   ---
 *
 * CreepTD PHP library Copyright (C) 2011 by
 * - Daniel Wirtz, virtunity media (http://virtunity.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * */
define("CREEPTD_PHP_VERSION", "1.0.5");
define("CREEPTD_BASE", dirname(__FILE__));
define("CREEPTD_MODE_ALLVSALL", 0);
define("CREEPTD_MODE_SENDNEXT", 1);
define("CREEPTD_MODE_SENDRANDOM", 2);
define("CREEPTD_MODE_TEAM2VS2", 3);
define("CREEPTD_MODE_SURVIVOR", 4);
define("CREEPTD_MODE_SURVIVORCOOP", 5);
define("CREEPTD_MAPSIZE_S16x16", "16x16");
define("CREEPTD_MAPSIZE_S16x32", "16x32");
define("CREEPTD_MAPSIZE_S32x16", "32x16");
define("CREEPTD_MAPSIZE_S32x32", "32x32");
require CREEPTD_BASE . "/CreepTD_Point.php";
require CREEPTD_BASE . "/CreepTD_Path.php";
require CREEPTD_BASE . "/CreepTD_PathSegment.php";
require CREEPTD_BASE . "/CreepTD_Map.php";
require CREEPTD_BASE . "/CreepTD_Language.php";

/**
 * CreepTD PHP library.
 * 
 * @author Daniel
 */
class CreepTD {

    /**
     * Parse CreepTD map data.
     * @param string $mapdata The map data to parse
     * @param array $warnings Parser warnings
     * @param string $message Validation message
     * @return CreepTD_Map The map object or null on error
     * @throws Exception If anything goes wrong
     */
    public static function &parseMapdata($mapdata, &$warnings = null, &$message = null) {
        $warnings = array();
        $message = null;
        $map = new CreepTD_Map();
        $message = $map->parse($mapdata, $warnings);
        return $map;
    }

    /**
     * Parse CreepTD map file.
     * @param string $mapfile The map file to parse
     * @param array $warnings Parser warnings
     * @param string $message Validation message
     * @return CreepTD_Map The map object or null on error
     * @throws Exception if anything goes wrong
     */
    public static function &parseMapfile($mapfile, &$warnings = null, &$message = null) {
        $warnings = array();
        $message = null;
        $map = new CreepTD_Map();
        if (substr($mapfile, -4) != '.map') {
        	throw(new Exception("Unable to open map file \"" . $mapfile . "\": Not a map file (*.map)"));
        }
        if (preg_match("/(map|coop)_([\\d\\w_\\-]+?)\\.map$/i", $mapfile, $matches)) {
            $map->setName(strtoupper($matches[2]));
        } else {
        	$map->setName(strtoupper(substr($mapfile, 0, -4)));
        }
        $mapdata = file_get_contents($mapfile);
        if ($mapdata === false) {
            throw(new Exception("Unable to open map file \"" . $mapfile . "\""));
        }
        $message = $map->parse($mapdata, $warnings);
        return $map;
    }
    
    /**
     * Parse a CreepTD map from a map archive.
     * @param string $archivefile The map archive file to open
     * @param array $warnings Parser warnings
     * @param string $message Validation message
     */
    public static function &parseMapArchive($archivefile, &$warnings = null, &$message = null) {
        $warnings = array();
        $message = null;
    	$archivefile = str_replace("\\", "/", $archivefile);
    	$zip = new ZipArchive;
    	if (substr($archivefile, -4) == ".zip") {
    		$pos = strrpos($archivefile, "/");
    		if ($pos === false) $pos = -1;
    		$lname = substr($archivefile, $pos+1, -4);
    		$zip = new ZipArchive;
	    	if (!$zip->open($archivefile)) {
	    		throw(new Exception("Unable to open map archive \"" . $archivefile . "\""));
	    	}
	    	$mapdata = $zip->getFromName($lname.".map");
	    	$zip->close();
	    	if (!$mapdata) {
	    		throw(new Exception("Unable to get \"".$lname.".map\" from map archive \"".$archivefile."\""));
	    	}
	    	$map = new CreepTD_Map();
	    	$map->setName(strtoupper($lname));
	    	$message = $map->parse($mapdata, $warnings);
	    	return $map;
    	} else {
    		throw(new Exception("Unable to open \"" . $archivefile . "\": Not a ZIP archive (*.zip)"));
    	}
    }
    
    /**
     * Package each map file in fromDir to an archive in toDir.
     * This method is basically used to update the old map file directory 
     * structures to the new packaged ones including validation and compilation.
     * @param string $fromDir Source directory
     * @param string $toDir Destination directory
     * @param string $previewDir Optional preview directory to copy the thumbnail to
     * @return array Packaged maps array (mapname => package filename)
     */
    public static function packageMaps($fromDir, $toDir, $previewDir = null) {
        $res = array();
        $dir = opendir($fromDir);
        if (!is_resource($dir)) return -1;
        $n=0; while (($filename = readdir($dir)) !== false) {
            if (preg_match("/^(map|coop)_([\\w]+)+\\.map$/", $filename)) {
                $mapname = substr($filename, 0, -4);
                $newmapname = substr($filename, strpos($mapname, "_")+1, -4);
                $toname = "$toDir/$newmapname.zip";
                
                // Parse, validate, compile
                try {
                    $map = CreepTD::parseMapfile($fromDir."/".$filename);
                    $mapdata = $map->compile();
                } catch (Exception $ex) {
                    echo "Unable to process map \"$mapname\": ".$ex."\n";
                    continue;
                }
                
                // Package
                @unlink($toname);
                $zip = new ZipArchive;
                $zip->open($toname, ZipArchive::CREATE);
                $zip->addFromString("$newmapname.map", $mapdata);
                $zip->addFile("$fromDir/images/$mapname.jpg", "$newmapname.jpg");
                $zip->close();
                if ($previewDir) {
                    copy("$fromDir/thumbnail/$mapname.jpg", "$previewDir/$newmapname.jpg");
                }
                $res[$mapname] = $toname;
            }
        }
        return $res;
    }
    
    /**
     * Get a list of valid map sizes for the given parameters.
     * @param type $mode The game mode played
     * @param type $numPlayers The number of players
     * @return array Valid map sizes
     */
    public static function getMapSizes($mode, $numPlayers) {
        $sizes = array();
        if ($mode == CREEPTD_MODE_TEAM2VS2) {
            $sizes[] = CREEPTD_MAPSIZE_S16x16;
        } else if ($mode == CREEPTD_MODE_SURVIVORCOOP) {
            if ($numPlayers == 2) {
                $sizes[] = CREEPTD_MAPSIZE_S16x32;
                $sizes[] = CREEPTD_MAPSIZE_S32x16;
            }
            $sizes[] = CREEPTD_MAPSIZE_S32x32;
        } else if ($mode == CREEPTD_MODE_SURVIVOR) {
            $sizes[] = CREEPTD_MAPSIZE_S16x16;
            if ($numPlayers <= 2) {
                $sizes[] = CREEPTD_MAPSIZE_S16x32;
                $sizes[] = CREEPTD_MAPSIZE_S32x16;
            }
            if ($numPlayers == 1) {
                $sizes[] = CREEPTD_MAPSIZE_S32x32;
            }
        } else {
            $sizes[] = CREEPTD_MAPSIZE_S16x16;
            if ($numPlayers == 2) {
                $sizes[] = CREEPTD_MAPSIZE_S16x32;
                $sizes[] = CREEPTD_MAPSIZE_S32x16;
            }
        }
        return $sizes;
    }
}

/**
 * Validate CreepTD map data.
 * 
 * This is a convenience function to simply test a map for validity. Use the
 * CreepTD class if you need the parsed CreepTD_Map object.
 * 
 * @param String $mapdata The map data to parse and validate
 * @param Array $warnings Parser warnings / errors
 * @return Boolean true on success, else false (see $warnings)
 * @see CreepTD::parseMap
 */
function creeptd_validate_mapdata($mapdata, &$warnings = null, $message = null) {
    try {
        $map = CreepTD::parseMapdata($mapdata, $warnings, $message);
        if (!$map) {
            return false;
        }
    } catch (Exception $ex) {
        $warnings[] = $ex;
        return false;
    }
    return true;
}

/**
 * Validate a CreepTD map file.
 * 
 * This is a convenience function to simply test a map for validity. Use the
 * CreepTD class if you need the parsed CreepTD_Map object.
 * 
 * @param String $mapfile Filename to parse and validate
 * @param Array $warnings Parser warnings / errors
 * @return Boolean true on success, else false (see $warnings)
 * @see CreepTD::parseMap
 */
function creeptd_validate_mapfile($mapfile, &$warnings = null, &$message = null) {
    try {
        $map = CreepTD::parseMapfile($mapfile, $warnings, $message);
        if (!$map) {
            return false;
        }
    } catch (Exception $ex) {
        $warnings[] = $ex;
        return false;
    }
    return true;
}
?>