<?php
/**
 * CreepTD is an Open Source Online Multiplayer TowerDefense game for 1-4
 * players playing on one map. It's the mother of all TD games.
 *
 *                      ---   www.creeptd.com   ---
 *
 * CreepTD PHP library Copyright (C) 2011 by
 * - Daniel Wirtz, virtunity media (http://virtunity.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * CreepTD Language Utility.
 * 
 * @author Daniel
 */
class CreepTD_Language {
    private $langDir;
    private $cacheDir;
    private $translations = array();
    private $languageKey = "en_US";
    private $languageName = "English (US)";
    
    /**
     * Set up website translations.
     * @param string $langDir The language file directory
     * @param string $cacheDir The cache directory
     */
    public function __construct($langDir, $cacheDir) {
        $this->langDir = $langDir;
        $this->cacheDir = $cacheDir;
        $langs = self::getBrowserLanguages();
        // Try to get the prefered language from the lang-Cookie
        if (isset($_COOKIE['ctdlang']) && preg_match("/^([a-zA-Z_]{1,8})$/", $_COOKIE['ctdlang'])) {
            unset($langs[$_COOKIE['ctdlang']]);
            $langs = array_merge(array($_COOKIE['ctdlang'] => 1.0), $langs);
            setcookie("ctdlang", $_COOKIE['ctdlang'], time()+2592000, "/"); // extend by 30d
        }
        // Try to get the language key from the current session if already set
        if (isset($_SESSION['ctdlang'])) {
            if ($this->includeLanguage($_SESSION['ctdlang'])) {
                return;
            }
        }
        // Else look for valid language
        foreach ($langs as $langKey => $langWeight) {
            if ($this->includeLanguage($langKey)) {
                $_SESSION['ctdlang'] = $langKey;
                break;
            }
        }
    }
    
    /**
     * Set prefered language.
     * @param string $langKey The language key to set
     */
    public function setPreferedLanguage($langKey) {
        if (preg_match("/^([a-zA-Z_]{1,8})$/", $langKey)) {
            unset($_SESSION['lang']);
            setcookie("ctdlang", $langKey, time()+2592000, "/"); // 30d
            return true;
        }
        return false;
    }
    
    /**
     * Include a language.
     * @param string $lang The language's key
     */
    private function includeLanguage($langKey) {
        $cachefile = $this->cacheDir."/language_".$langKey.".php";
        $langfile = $this->langDir."/".$langKey.".txt";
        $lang = null;
        // Check if there is a recent cached version of the language
        if (file_exists($cachefile) && filemtime($cachefile) >= filemtime($langfile)) {
            include $cachefile;
            if (isset($__LANG)) {
                $lang = $__LANG;
            }
        }
        // Else check if there is a plain language file and if it is,
        // create a new cache file
        if (file_exists($langfile)) {
            $lang = self::parseLanguageFile(dirname($langfile), basename($langfile));
            $lang['includes'] = array_keys($lang['includes']);
            $code = '<?php $__LANG = '.var_export($lang, true).'; ?'.'>';
            $fp = fopen($cachefile, "w");
            if ($fp) {
                flock($fp, LOCK_EX | LOCK_NB, $wb);
                if (!$wb) {
                    fwrite($fp, $code);
                    flock($fp, LOCK_UN);
                    fclose($fp);
                }
            }
        }
        
        // Finally build the translation index
        if ($lang !== null) {
            $this->languageKey = $langKey;
            $this->languageName = $lang['name'];
            foreach ($lang['includes'] as $includedLangKey) {
                $this->includeLanguage($includedLangKey);
            }
            $this->translations = array_merge($this->translations, $lang['translations']);
            return true;
        }
        return false;
    }
    
    /**
     * Get the current language key.
     * @return string Language key
     */
    public function getLanguageKey() {
        return $this->languageKey;
    }
    
    /**
     * Get language key for usage in HTML tag.
     * @return string Language key for HTML tag
     */
    public function getHtmlTagLang() {
        return str_replace("_", "-", $this->languageKey);
    }
    
    /**
     * Translate a String.
     * @param string $s String to translate
     * @param array $args Optional replace argument
     */
    public function translate($s, $args = array()) {
        $s = str_replace("\r", "", $s);
        $s = str_replace("\n", "\\n", $s);
        $s = str_replace("\t", "\\t", $s);
        if (isset($this->translations[$s])) {
            $s = $this->translations[$s];
        }
        $s = str_replace("\\n", "\n", $s);
        $s = str_replace("\\t", "\t", $s);
        foreach ($args as $key => $val) {
            $s = str_replace("%$key%", $val, $s);
        }
        return $s;
    }

    /**
     * Get a storted hash of accepted browser languages.
     * @return array Sorted hash of accepted languages
     */
    public static function getBrowserLanguages() {
        $langs = array();
        preg_match_all('/(([a-z]{1,8})(-[a-z]{1,8})?)\s*(;\s*q\s*=\s*(1|0\.[0-9]+))?/i', $_SERVER['HTTP_ACCEPT_LANGUAGE'], $matches);
        if (!empty($matches[1])) {
            foreach ($matches[2] as $i => $lang) {
                $loc = ($matches[3][$i] != "") ? substr($matches[3][$i], 1) : null;
                $key = ($loc != null) ? $lang . "_" . strtoupper($loc) : $lang;
                $val = ($matches[5][$i] == "") ? 1.0 : floatval($matches[5][$i]);
                $langs[$key] = $val;
            }
            arsort($langs, SORT_NUMERIC);
        }
        return $langs;
    }

    /**
     * Parse all Java sources in the given project directory.
     * @param string $dirname Project directory
     * @return array Hash of used translations and their locations
     */
    public static function parseJavaSources($dirname) {
        $base = str_replace("\\", "/", realpath($dirname . "/src/java/com/creeptd"));
        $translations = self::parseSources($base, "java");
        foreach ($translations as $key => $locations) {
            foreach ($locations as $i => $filename) {
                $translations[$key][$i] = "com/creeptd/" . substr($filename, strlen($base) + 1);
            }
        }
        return $translations;
    }

    /**
     * Parse all PHP sources in the given project directory.
     * @param string $dirname Project directory
     * @return array Hash of used translations and their locations
     */
    public static function parsePHPSources($dirname) {
        $base = str_replace("\\", "/", realpath($dirname . "/www"));
        $translations = self::parseSources($base, "php");
        foreach ($translations as $key => $locations) {
            foreach ($locations as $i => $filename) {
                $translations[$key][$i] = substr($filename, strlen($base) + 1);
            }
        }
        return $translations;
    }

    /**
     * Parse the given source directory and return an array of used translation
     * strings.
     * @param string $dirname Directory to search through
     * @param boolean $recursive true to enable recursion into sub directories, else false
     * @param array $result Internally used result reference
     * @return array Hash of used translations and their locations
     */
    private static function parseSources($dirname, $ext, $recursive = true, &$result = array()) {
        $dirname = str_replace("\\", "/", $dirname);
        $dir = opendir($dirname);
        if ($dir === false) {
            return null;
        }
        while (($filename = readdir($dir)) !== false) {
            if ($filename == "." || $filename == "..") {
                continue;
            }
            if ($recursive && is_dir($dirname . "/" . $filename)) {
                self::parseSources($dirname . "/" . $filename, $ext, $recursive, $result);
            } else if (substr($filename, -(strlen($ext) + 1)) == "." . $ext) {
                self::parseSource($dirname . "/" . $filename, $result);
            }
        }
        closedir($dir);
        return $result;
    }

    /**
     * Parse a source file and look for used translations.
     * @param string $filename The file to parse
     * @return array List of used translations
     */
    private static function parseSource($filename, &$result) {
        $filename = str_replace("\\", "/", realpath($filename));
        $raw = file_get_contents($filename);
        while (($pos = strpos($raw, "_(\"")) !== false) {
            $raw = substr($raw, $pos + 3);
            $t = "";
            $esc = false;
            $i = 0;
            while (true) {
                if (strlen($raw) <= $i) {
                    break 2; // EOF, not a translation
                }
                $chr = $raw{$i};
                if (!$esc) {
                    if ($chr == '\\') {
                        $esc = !$esc;
                    } else if ($chr == '"') {
                        if (!isset($result[$t])) {
                            $result[$t] = array();
                        }
                        if (!in_array($filename, $result[$t])) {
                            $result[$t][] = $filename;
                        }
                        $raw = substr($raw, $i + 1);
                        break;
                    } else {
                        if ($chr == "\n") {
                            $t .= "\\n";
                        } else if ($chr == "\t") {
                            $t .= "\\t";
                        } else if ($chr == "\r") {
                            // Skip
                        } else {
                            $t .= $chr;
                        }
                    }
                } else {
                    $esc = false;
                    if ($chr == '"') {
                        $t .= $chr;
                    } else {
                        $t .= "\\$chr";
                    }
                }
                $i++;
            }
        }
    }

    /**
     * Update all language files to match the given translation version.
     * @param string $dirname The language files' directory
     * @param array $translations Array of translations
     * @return array Translation report array or null on error
     */
    public static function updateLanguageFiles($dirname, $translations) {
        $res = array();
        $dir = opendir($dirname);
        if ($dir === false) {
            return null;
        }
        $messages = array();
        while (($filename = readdir($dir)) !== false) {
            if (substr($filename, -4) == ".txt") {
                $lang = substr($filename, 0, -4);
                $msgs = self::updateLanguageFile($dirname, $filename, $translations);
                $messages = array_merge($messages, $msgs);
            }
        }
        closedir($dir);
        return $messages;
    }

    /**
     * Update a language file to match the currently used translations.
     * @param string $dirname The directory where language files are located
     * @param string $filename The language file's name
     * @param array $translations Array of translations
     */
    public static function updateLanguageFile($dirname, $filename, $translations) {
        $res = self::parseLanguageFile($dirname, $filename);
        $name = $res['name'];
        $includes = $res['includes'];
        $used = $res['translations'];
        $data = "$name\r\n\r\n";
        $messages = array();
        $included = array();
        if (count($includes) > 0) {
            foreach ($includes as $lang => $trans) {
                $data .= "@import $lang\r\n";
                foreach ($trans as $key => $val) {
                    $included[$key] = $val;
                }
            }
            $data .= "\r\n";
        }
        foreach ($used as $key => $val) {
            if (!isset($translations[$key])) {
                $messages[] = "Translation \"$key\" is no longer used in \"$filename\"";
            }
        }
        foreach ($translations as $key => $fnames) {
            if (!isset($used[$key]) && !isset($included[$key])) {
                $messages[] = "Translation \"$key\" is untranslated in \"$filename\"";
                if (count($includes) == 0) {
                    $data .= "$key\r\n\r\n\r\n";
                }
            } else {
                if (!isset($included[$key]) || $included[$key] != $used[$key]) {
                    $data .= "$key\r\n" . $used[$key] . "\r\n\r\n";
                }
            }
        }
        $fp = fopen($dirname."/".$filename, "w");
        if ($fp) {
            flock($fp, LOCK_EX);
            fwrite($fp, $data);
            flock($fp, LOCK_UN);
            fclose($fp);
        } else {
            $messages[] = "Unable to open language file \"$filename\" for writing";
        }
        return $messages;
    }

    /**
     * Parse a Java source language file.
     * @param string $dirname The directory where languages are located
     * @param string $filename The language file's name
     * @return array Hash of author, includes and translations
     */
    public static function parseLanguageFile($dirname, $filename) {
        $res = array(
            "name" => null,
            "translations" => array(),
            "includes" => array()
        );
        $fp = fopen($dirname . "/" . $filename, "r");
        if (!$fp)
            return null;
        $key = null;
        $lineNo = 0; while (!feof($fp)) {
            $line = rtrim(fgets($fp));
            $lineNo++;
            if (strlen($line) > 0 && ($line{0} == ' ' || $line{0} == "\t")) continue;
            if ($lineNo == 1) {
                $res['name'] = trim(str_replace("﻿", "", $line));
                continue;
            }
            if (preg_match("/^@import\\s+([a-zA-Z0-9_]+)$/", $line, $matches)) {
                $sublang = $matches[1];
                $subres = self::parseLanguageFile($dirname, $sublang . ".txt");
                $res['includes'][$sublang] = $subres;
            } else {
                if ($key === null) {
                    if (strlen($line) > 0) {
                        $key = $line;
                    }
                } else {
                    if (strlen($line) > 0) {
                        $res['translations'][$key] = $line;
                    }
                    $key = null;
                }
            }
        }
        fclose($fp);
        return $res;
    }

}
?>