<?php
/**
 * CreepTD is an Open Source Online Multiplayer TowerDefense game for 1-4
 * players playing on one map. It's the mother of all TD games.
 *
 *                      ---   www.creeptd.com   ---
 *
 * CreepTD PHP library Copyright (C) 2011 by
 * - Daniel Wirtz, virtunity media (http://virtunity.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/
define("CREEPTD_GRID_WIDTH_MIN", 16);
define("CREEPTD_GRID_WIDTH_MAX", 32);
define("CREEPTD_GRID_HEIGHT_MIN", CREEPTD_GRID_WIDTH_MIN);
define("CREEPTD_GRID_HEIGHT_MAX", CREEPTD_GRID_WIDTH_MAX);

/**
 * A CreepTD map.
 * 
 * @author Daniel
 */
class CreepTD_Map {

    private $name = null;
    private $path = null;
    private $holyland = null;
    private $width = 0;
    private $height = 0;
    private $author = null;
    
    /**
     * Create a new map instance.
     */
    public function __construct() {
    }
    
    /**
     * Get the map's name.
     * @return string The map's name or null if not defined
     */
    public function getName() {
        return $this->name;
    }
    
    /**
     * Set the map's name.
     * @param string $name The name to set
     */
    public function setName($name) {
        $this->name = $name;
    }
    
    /**
     * Get the map author's name.
     * @return string The map author's name or null if undefined
     */
    public function getAuthor() {
        return $this->author;
    }
    
    /**
     * Set the map author's name.
     * @param string $author The map author's name
     */
    public function setAuthor($author) {
        $this->author = $author;
    }
    
    /**
     * Get the map's checksum.
     */
    public function getChecksum() {
    	return $this->checksum;
    }

    /**
     * Parse and validate map data.
     * @param string $mapdata Map data
     * @param array $warnings List of parser warnings
     * @param boolean validate Whether to also validate the map data or not
     * @return mixed If validate=true is given, the final validation message
     *                will be returend, otherwise true is returned
     */
    public function parse($mapdata, array &$warnings, $validate = true) {
        if (empty($warnings))
            $warnings = array();

        $path = new CreepTD_Path();
        $holyland = array();

        // Annotations
        $author = null;
        $mapWidth = null;
        $mapHeight = null;
        $parseAnnotations = true;
        
        // Whether strict mode is enabled or not (default: off)
        $strictMode = false;
        
        // Map checksum
        $mapdata = str_replace("\r", "", $mapdata);
        $checksum = md5($mapdata);
        
        $lines = explode("\n", $mapdata);
        for ($lineNo = 1; $lineNo <= count($lines); $lineNo++) {
            $line = rtrim($lines[$lineNo-1]);
            $ci = strpos($line, "#");
            if ($ci !== false) { // Strip comments
                $line = rtrim(substr($line, 0, $ci));
            }
            if (strlen($line) == 0) { // Empty line
                continue;
            }
            // Path definition
            if (preg_match("/^(\\s*)([\\-]?[\\d]+)\\s*,\\s*([\\-]?[\\d]+)(\\s+.+|)$/", $line, $matches)) {
                $parseAnnotations = false;
                $ug = strlen($matches[1]) > 0;
                $x = intval($matches[2]);
                $y = intval($matches[3]);
                $start = false;
                $end = false;
                $params = trim($matches[4]);
                if (strlen($params) >= 5 && substr($params, 0, 5) == "start" && (strlen($params) == 5 || $params{5} == " " || $params{5} == "\t")) {
                    $params = rtrim(substr($params, 5));
                    $start = true;
                } else if (strlen($params) >= 3 && substr($params, 0, 3) == "end" && (strlen($params) == 3 || $params{3} == " " || $params{3} == "\t")) {
                    $params = rtrim(substr($params, 3));
                    $end = true;
                }
                $segment = new CreepTD_PathSegment($x, $y, $start, $end, $ug, $lineNo);
                if ($mapWidth == null) {
                    if ($strictMode) {
                        throw(new Exception("Parser failed at line ".$lineNo.": In strict mode, the @size annotation is required"));
                    } else {
                        $mapWidth = CREEPTD_GRID_WIDTH_MIN;
                        $mapHeight = CREEPTD_GRID_HEIGHT_MIN;
                        $warnings[] = "No @size annotation found at line ".$lineNo.": Assuming ".$mapWidth."x".$mapHeight;
                    }
                }
                if ($this->testOOB($x, $y, $ug, $mapWidth, $mapHeight)) {
                    throw(new Exception($segment." is out of bounds at line ".$lineNo));
                }
                if ($params != null && strlen($params) > 0) {
                    // Jumps?
                    $gi = strpos($params, "goto ");
                    if ($gi !== false) {
                        $labels = explode(",", substr($params, $gi+5));
                        for ($i = 0; $i < count($labels); $i++) {
                            $labels[$i] = trim($labels[$i]);
                            if (strlen($labels[$i]) > 0) {
                                if ($segment->isEnd()) {
                                    throw(new Exception("Invalid end segment declaration \"".$line."\" found at line ".$lineNo.": End segments cannot contain jump labels"));
                                }
                                $segment->addJumpLabel($labels[$i]);
                            }
                        }
                        $params = trim(substr($params, 0, $gi));
                    }
                    // Label?
                    if (strlen($params) > 0) {
                        if (strpos($params, " ") !== false) {
                            // In most cases this should be a typo -> exit.
                            throw(new Exception("Invalid whitespaced label \"".$params."\" found at line ".$lineNo));
                        }
                        $segment->setLabel($params);
                    }
                }
                $path->addSegment($segment);
                unset($segment);
                continue;

            // Holyland defintion
            } else if (preg_match("/^(\\d+)\\s*;\\s*(\\d+)$/", $line, $matches)) {
                $parseAnnotations = false;
                $x = intval($matches[1]);
                $y = intval($matches[2]);
                if ($mapWidth == null) {
                    if ($strictMode) {
                        throw(new Exception("Parser failed at line ".$lineNo.": In strict mode, the @size annotation is required"));
                    } else {
                        $mapWidth = CREEPTD_GRID_WIDTH_MIN;
                        $mapHeight = CREEPTD_GRID_HEIGHT_MIN;
                        $warnings[] = "No @size annotation found at line ".$lineNo.": Assuming ".$mapWidth."x".$mapHeight;
                    }
                }
                if ($this->testOOB($x, $y, false, $mapWidth, $mapHeight)) {
                    throw(new Exception("Holyland (".$x.",".$y.") is out of bounds at line ".$lineNo));
                }
                $p = new CreepTD_Point($x, $y);
                $holyland[] =& $p;
                unset($p);
                continue;
            }
            
            if ($parseAnnotations) {
                $line = trim($line); // Trim leading white spaces
                
                // Deprecated map name hint
                if (!$strictMode && preg_match("/map_([\\w\\d_\\-]+)\\.map$/i", $line, $matches) && $this->getName() == null) {
                    $this->setName(strToUpper($matches[1]));
                    $warnings[] = "Deprecated name hint found at line ".$lineNo.": Assuming \"".$this->getName()."\" as map name anyway (was undefined)";
                    continue;
                }

                // Author annotation
                if (substr($line, 0, 8) == "@author ") {
                    $authorName = trim(substr($line, 8));
                    if (strlen($authorName) > 0) {
                        if (strlen($authorName) > 30) $authorName = substr($authorName, 0,30);
                        $author = $authorName;
                        continue;
                    }
                }

                // Explicitly activating strict mode
                if ($line == "@strict") {
                    $strictMode = true;
                    continue;
                }
                
                // Size annotation
                if (preg_match("/^@size\\s+(".CREEPTD_GRID_WIDTH_MIN."|".CREEPTD_GRID_WIDTH_MAX.")\\s*x\\s*(".CREEPTD_GRID_HEIGHT_MIN."|".CREEPTD_GRID_HEIGHT_MAX.")$/", $line, $matches)) {
                    $mapWidth = intval($matches[1]);
                    $mapHeight = intval($matches[2]);
                    continue;
                }
            }
            
            // Malformed line
            if ($strictMode || !$parseAnnotations) {
                throw(new Exception("Unable to parse map: Parse error at line " . $lineNo . ": " . $line));
            } else {
                $warnings[] = "Ignoring deprecated or malformed definition at line " . $lineNo . ": " . $line;
            }
        }
        if ($validate) {
            $message = null;
            try {
                $message = $path->validate($strictMode, $warnings);
            } catch (Exception $ex) {
                throw(new Exception("Unable to parse map: Validation failed: " . $ex));
            }
        }
        $this->path =& $path;
        $this->holyland =& $holyland;
        $this->width = $mapWidth;
        $this->height = $mapHeight;
        $this->author = $author;
        $this->checksum = $checksum;
        return ($validate) ? "Map validated: ".$message : true;
    }
    
    /**
     * Test if the map path has been validated.
     * @return boolean true if validated, else false
     */
    public function isValidated() {
    	return $this->path != null && $this->path->isValidated();
    }
    
    /**
     * Test if a coordinate is out of bounds.
     * @param int $x X coordinate to test
     * @param int $y Y coordinate to test
     * @param boolean $ug true if underground, else false
     * @param int $maxWidth The maximum map width
     * @param int $maxHeight The maximum map height
     * @return boolean true if out of bounds, else false
     */
    public static function testOOB($x, $y, $ug, $maxWidth, $maxHeight) {
        if ($ug) {
            return $x < -1 || $x > $maxWidth || $y < -1 || $y > $maxHeight;
        }
        return $x < 0 || $x > $maxWidth-1 || $y < 0 || $y > $maxHeight-1;
    }
    
    /**
     * Get the map's width.
     * @return int CREEPTD_GRID_WIDTH_MIN or CREEPTD_GRID_WIDTH_MAX, 0 if not yet calculated
     */
    public function getWidth() {
        return $this->width;
    }
    
    /**
     * Get the map's height.
     * @return int CREEPTD_GRID_HEIGHT_MIN or CREEPTD_GRID_HEIGHT_MAX, 0 if not yet calculated
     */
    public function getHeight() {
        return $this->height;
    }
    
    /**
     * Get the map's size as a string.
     * @return One of "16x16", "16x32", "32x16", "32x32" or null if not yet known
     */
    public function getSizeString() {
    	if ($this->width > 0 && $this->height > 0) {
	    	return $this->width."x".$this->height;
    	}
    	return null;
    }

    /**
     * Get the map's path.
     * @return CreepTD_Path The map's path or null if not validated
     */
    public function getPath() {
        return $this->path;
    }

    /**
     * Get a list of all holyland coordinates.
     * @return array List of all holyland coordinates or null if not validated
     */
    public function getHolyland() {
        return $this->holyland;
    }
    
    /**
     * Compile the map elements to map file data.
     * @return string The map file data or null if not valid
     */
    public function compile() {
        if ($this->path == null || !$this->path->isValidated()) {
            return null;
        }
        $s = "";
        $name = ($this->getName()) ? $this->getName() : "%UNNAMED%";
        $s .= "# Map ".$name."\r\n";
        $s .= "# compiled by CreepTD PHP library ".CREEPTD_PHP_VERSION."\r\n# at ".date("r")."\r\n\r\n";
        if ($this->getAuthor() != null) {
            $s .= "@author ".$this->getAuthor()."\r\n";
        }
        $s .= "@strict\r\n";
        $s .= "@size ".$this->getWidth()."x".$this->getHeight()."\r\n\r\n";
        $s .= "# Path segments\r\n";
        $segments =& $this->path->getSegments();
        for ($i=0; $i<count($segments); $i++) {
            $p =& $segments[$i];
            $s .= $p->compile()."\r\n";
            if ($p->isEnd() || count($p->getJumpLabels()) > 0) {
                $s .= "\r\n";
            }
            unset($p);
        }
        if (count($this->holyland) > 0) {
            $s .= "# Holyland coordinates\r\n";
            for ($i=0; $i<count($this->holyland); $i++) {
                $p =& $this->holyland[$i];
                $s .= $p->getX().";".$p->getY()."\r\n";
                unset($p);
            }
        } else {
            $s .= "# No holyland coordinates defined\r\n";
        }
        return $s;
    }
}
?>