<?php
/**
 * CreepTD is an Open Source Online Multiplayer TowerDefense game for 1-4
 * players playing on one map. It's the mother of all TD games.
 *
 *                      ---   www.creeptd.com   ---
 *
 * CreepTD PHP library Copyright (C) 2011 by
 * - Daniel Wirtz, virtunity media (http://virtunity.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * A CreepTD path.
 * 
 * @author Daniel
 */
class CreepTD_Path {

    private $segments = array();
    private $starts = array();
    private $ends = array();
    private $validated = false;
    private $min_length = -1;
    private $max_length = -1;

    public function __construct() {
        
    }

    /**
     * Get a list of all PathSegments.
     * @return array List of PathSegments
     */
    public function &getSegments() {
        return $this->segments;
    }

    /**
     * Add a PathSegment.
     * @param CreepTD_PathSegment $segment The segment to add
     */
    public function addSegment(&$segment) {
        $this->segments[] = & $segment;
        $this->validated = false;
    }

    /**
     * Get a List of all start PathSegments.
     * @return array List of all start segments
     */
    public function &getStarts() {
        return $this->starts;
    }

    /**
     * Get the number of starts.
     * @return int Number of starts
     */
    public function getNumStarts() {
        return count($this->starts);
    }

    /**
     * Get a list of all end PathSegments.
     * @return array List of all end segments
     */
    public function &getEnds() {
        return $this->ends;
    }
    
    /**
     * Get the number of ends.
     * @return int Number of ends
     */
    public function getNumEnds() {
    	return count($this->ends);
    }
    
    /**
     * Get the minimum path length.
     * @return int Minimum path length or -1 if not yet calculated
     */
    public function getMinLength() {
    	return $this->min_length;
    }
    
    /**
     * Get the maximum path length.
     * @return int Maximum path length or -1 if not yet calculated
     */
    public function getMaxLength() {
		return $this->max_length;
    }
    
    /**
     * Get the number of path splittings.
     * @return int Number of splittings
     */
    public function getNumSplittings() {
    	$num = 0;
    	for ($i=0; $i<count($this->segments); $i++) {
    		$p =& $this->segments[$i];
    		$n = count($p->getSuccessors());
    		if ($n > 1) {
    			$num += $n-1;
    		}
    		unset($p);
		}
		return $num;
    }
    
    /**
     * Get the number of path joins.
     * @return int Number of joins
     */
    public function getNumJoins() {
    	$num = 0;
    	for ($i=0; $i<count($this->segments); $i++) {
    		$p =& $this->segments[$i];
    		$n = count($p->getPredecessors());
    		if ($n > 1) {
    			$num += $n-1;
    		}
    		unset($p);
		}
		return $num;
    }
    
    /**
     * Get the number of overground segments.
     * @return int Number of overground segments
     */
    public function getNumOvergrounds() {
    	$num = 0;
    	for ($i=0; $i<count($this->segments); $i++) {
    		if (!$this->segments[$i]->isUnderground()) {
    			$num++;
    		}
    	}
    	return $num;
    }
    
    /**
     * Get the number of underground segments.
     * @return int Number of underground segments
     */
    public function getNumUndergrounds() {
    	$num = 0;
    	for ($i=0; $i<count($this->segments); $i++) {
    		if ($this->segments[$i]->isUnderground()) {
    			$num++;
    		}
    	}
    	return $num;
    }
    
    /**
     * Validate Path.
     * @param boolean $strictMode true to validate in strict mode
     * @param array $warnings Warning array reference
     * @return string Validation message
     * @throws Exception If anything goes wrong
     */
    public function validate($strictMode = false, &$warnings = null) {
        if (empty($warnings)) $warnings = array();
        
        $this->validated = false;
        $this->starts = array();
        $this->ends = array();
        
        // Check if there are any segments
        if (empty($this->segments)) {
            throw(new Exception("Path is empty"));
        }
        // Build index and initialize segments for (re-)validation
        $labels = array();
        $usedLabels = array();
        for ($i=0; $i<count($this->segments); $i++) {
            $p =& $this->segments[$i];
            $p->clearSuccessors(); // If revalidating...
            $p->clearPredecessors();
            $p->setDepth(PHP_INT_MAX);
            if ($p->isStart()) {
                $this->starts[] =& $p;
            } else if ($p->isEnd()) {
                $this->ends[] =& $p;
            } else if ($p->getLabel() != null) {
                if (isset($labels[$p->getLabel()])) {
                    throw(new Exception("Duplicate label \"".$p->getLabel()."\" detected at line ".$p->getLineNo()));
                }
                $labels[$p->getLabel()] =& $p;
            }
            unset($p);
        }
        // Check starts and fall back to old format if missing
        if (!$this->segments[0]->isStart()) {
            $start =& $this->segments[0];
            if ($strictMode) {
                throw(new Exception("Undeclared start segment \"".$start."\" found at line ".$start->getlineNo()));
            }
            $start->setStart(true);
            $this->starts[] =& $start;
            $warnings[] = "Undeclared start segment found at line ".$start->getlineNo().": Falling back to \"".$start."\"";
            unset($start);
        }
        // Parse jumps and create segment tree
        $prev = null;
        for ($i=0; $i<count($this->segments); $i++) {
            $p =& $this->segments[$i];
            if ($prev != null) { // See below
                $prev->addSuccessor($p);
                $p->addPredecessor($prev);
                unset($prev);
                $prev = null;
            }
            if (count($p->getJumpLabels()) > 0) {
                $jumpLabels =& $p->getJumpLabels();
                for ($j=0; $j<count($jumpLabels); $j++) {
                    $jumpLabel = $jumpLabels[$j];
                    if (!isset($labels[$jumpLabel])) {
                        throw(new Exception("Jump label \"".$jumpLabel."\" is undefined at line ".$p->getLineNo()));
                    }
                    $usedLabels[] = $jumpLabel;
                    $suc =& $labels[$jumpLabel];
                    $p->addSuccessor($suc);
                    $suc->addPredecessor($p);
                }
                unset($jumpLabels);
            } else if (!$p->isEnd()) {
                $prev =& $p; // Append next point to this one (walk path by lines)
            }
            unset($p);
        }
        // Check for unused labels
        foreach ($labels as $jumpLabel => $p) {
            if (!in_array($jumpLabel, $usedLabels)) {
                $msg = "Label \"".$jumpLabel."\" declared at line ".$p->getLineNo()." is not used";
                if ($strictMode) {
                    throw(new Exception($msg));
                } else {
                    $warnings[] = $msg;
                }
            }
        }
        // Check if path is correctly terminated, else fall back
        if (!$strictMode) {
            for ($i=0; $i<count($this->segments); $i++) {
                $p =& $this->segments[$i];
                if (!$p->isEnd() && count($p->getSuccessors()) == 0) {
                    $p->setEnd(true);
                    $this->ends[] =& $p;
                    $warnings[] = "Loose end segment found at line ".$p->getLineNo().": Falling back to \"".$p."\"";
                }
                unset($p);
            }
        }
        if (count($this->ends) == 0) {
            throw(new Exception("Unable to find any valid end segments"));
        }
        // Calculate depth
        for ($i=0; $i<count($this->ends); $i++) {
            $p =& $this->ends[$i];
            $this->calcDepth($p, PHP_INT_MAX-1);
            unset($p);
        }
        // Get minimum depth
        $minDepth = PHP_INT_MAX;
        for ($i=0; $i<count($this->segments); $i++) {
            $p =& $this->segments[$i];
            if ($p->getDepth() < $minDepth) {
                $minDepth = $p->getDepth();
            }
            unset($p);
        }
        // Find infinite loops and normalize depth values
        for ($i=0; $i<count($this->segments); $i++) {
            $p =& $this->segments[$i];
            if ($p->getDepth() == PHP_INT_MAX) {
                throw(new Exception("Infinite loop detected at line ".$p->getLineNo()));
            }
            $p->setDepth($p->getDepth()-$minDepth+1);
            unset($p);
        }
        $this->validated = true;
        list($min, $max) = $this->calcLength($this->starts);
        $this->min_length = $min;
        $this->max_length = $max;
        return count($this->segments)." segments, ".count($this->starts)." starts, ".count($this->ends)." ends";
    }
    
    /** 
     * Calculate depth of given segment.
     * {p1,p2} with p1.depth > p2.depth => p1 nearer to the end
     * @param CreepTD_PathSegment $p The current PathSegment
     * @param int $depth The current depth
     */
    private function calcDepth(CreepTD_PathSegment $p, $depth) {
        if ($p->getDepth() <= $depth) {
            return; // Remember shortest depth to end
        }
        $p->setDepth($depth);
        $pp =& $p->getPredecessors();
        for ($i=0; $i<count($pp); $i++) {
            $this->calcDepth($pp[$i], $depth-1);
        }
    }
    
    /**
     * Calculate minimum and maximum path length.
     * Please note: Its absolutely neccessary to test the path for endless loops before calculating its lenght.
     * @param CreepTD_PathSegment $p The PathSegment to start at
     * @return array Minimum (index 0) and maximum (index 1) path mength
     */
    private function calcLength(&$ps) {
    	if (empty($ps)) return array(0, 0);
    	$min = PHP_INT_MAX;
    	$max = 0;
    	for ($i=0; $i<count($ps); $i++) {
    		$p =& $ps[$i];
    		list($submin, $submax) = $this->calcLength($p->getSuccessors());
    		if ($submin+1 < $min) $min = $submin+1;
    		if ($submax+1 > $max) $max = $submax+1;
    		unset($p);
    	}
    	return array($min, $max);
    }
    
    /**
     * Test if the Path has been successfully validated.
     * @return boolean true if successfully validated, else false
     */
    public function isValidated() {
        return $this->validated;
    }

}

?>