<?php
/**
 * CreepTD is an Open Source Online Multiplayer TowerDefense game for 1-4
 * players playing on one map. It's the mother of all TD games.
 *
 *                      ---   www.creeptd.com   ---
 *
 * CreepTD PHP library Copyright (C) 2011 by
 * - Daniel Wirtz, virtunity media (http://virtunity.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 **/

/**
 * A CreepTD path segment.
 * 
 * @author Daniel
 */
class CreepTD_PathSegment {

    private $x;
    private $y;
    private $start;
    private $end;
    private $underground;
    private $lineNo;
    private $label = null;
    private $jumpLabels = array();
    private $successors = array();
    private $predecessors = array();
    private $depth = PHP_INT_MAX;

    /**
     * Create a new PathSegment.
     * @param int $x The segment's X coordinate
     * @param int $y The segment's Y coordinate
     * @param boolean $start true if a start, else false
     * @param boolean $end true if an end, else false
     * @param boolean $underground true if an underground segment, else false
     * @param int $lineNo The parser's current line number for debug outputs
     */
    public function __construct($x, $y, $start, $end, $underground, $lineNo) {
        $this->x = $x;
        $this->y = $y;
        $this->start = $start;
        $this->end = $end;
        $this->underground = $underground;
        $this->lineNo = $lineNo;
    }

    /**
     * Get the segment's X coordinate.
     * @return int X coordinate
     */
    public function getX() {
        return $this->x;
    }

    /**
     * Get the segment's Y coordinate.
     * @return int Y coordinate
     */
    public function getY() {
        return $this->y;
    }

    /**
     * Test if this segment is a start.
     * @return boolean true if a start, else false
     */
    public function isStart() {
        return $this->start;
    }
    
    /**
     * Set if this segment is a start.
     * @param boolean $start true if a start, else false
     */
    public function setStart($start) {
        $this->start = $start;
    }

    /**
     * Test if this segment is an end.
     * @return boolean true if an end, else false
     */
    public function isEnd() {
        return $this->end;
    }
    
    /**
     * Set if this segment is an end.
     * @param boolean $end true if an end, else false
     */
    public function setEnd($end) {
        $this->end = $end;
    }

    /**
     * Test if this is an underground segment.
     * @return boolean true if an underground segment, else false
     */
    public function isUnderground() {
        return $this->underground;
    }

    /**
     * Get the parser's line number.
     * @return int The parser's line number
     */
    public function getLineNo() {
        return $this->lineNo;
    }
    
    /**
     * Get the segment's label.
     * @return string The segment's label or null if none
     */
    public function getLabel() {
        return $this->label;
    }

    /**
     * Set the segment's label.
     * @param string $label The label to set
     */
    public function setLabel($label) {
        $this->label = $label;
    }

    /**
     * Get the segment's jump labels.
     * @return array List of jump labels
     */
    public function getJumpLabels() {
        return $this->jumpLabels;
    }

    /**
     * Add a jump label to this segment.
     * @param string $jumpLabel The jump label to add
     */
    public function addJumpLabel($jumpLabel) {
        $this->jumpLabels[] = $jumpLabel;
    }
    
    /**
     * Clear all jump labels.
     */
    public function clearJumpLabels() {
        $this->jumpLabels = array();
    }

    /**
     * Get a list of this segment's successors.
     * @return array List of this segment's successors
     */
    public function &getSuccessors() {
        return $this->successors;
    }
    
    /**
     * Add a successor to this segment.
     * @param CreepTD_PathSegment $successor The successor to add
     */
    public function addSuccessor(CreepTD_PathSegment &$successor) {
        $this->successors[] =& $successor;
    }
    
    /**
     * Clear this segments successors.
     */
    public function clearSuccessors() {
        $this->successors = array();
    }
    
    /**
     * Get a list of this segment's predecessors.
     * @return array List of this segment's predecessors
     */
    public function &getPredecessors() {
        return $this->predecessors;
    }
    
    /**
     * Add a predecessor to this segment.
     * @param CreepTD_PathSegment $predecessor The predecessor to add
     */
    public function addPredecessor(CreepTD_PathSegment &$predecessor) {
        $this->predecessors[] =& $predecessor;
    }
    
    /**
     * Clear this segment's predecessors.
     */
    public function clearPredecessors() {
        $this->predecessors = array();
    }
    
    /**
     * Get this segment's depth value.
     * @return int This segment's depth value
     */
    public function getDepth() {
        return $this->depth;
    }
    
    /**
     * Set this segment's depth value.
     * @param int $depth The depth value to set
     */
    public function setDepth($depth) {
        $this->depth = $depth;
    }
    
    /**
     * Compile the segment's data to a Map file declaration.
     * @return string The compiled segment
     */
    public function compile() {
        $s = "";
        if ($this->isUnderground()) {
            $s .= " ";
        }
        $s .= $this->getX().",".$this->getY();
        if ($this->isStart()) {
            $s .= " start";
        } else if ($this->isEnd()) {
            $s .= " end";
        } else if ($this->getLabel() != null) {
            $s .= " ".$this->getLabel();
        } else if (count($this->jumpLabels) > 0) {
            $s .= " goto ";
            for ($i=0; $i<count($this->jumpLabels); $i++) {
                if ($i>0) $s.= ",";
                $s .= $this->jumpLabels[$i];
            }
        }
        return $s;
    }
    
    public function __toString() {
        return $this->compile();
    }
}

?>